@@ -14,6 +14,7 @@ import time
 import boto3
 import requests
 import multiprocessing
+import matplotlib.image as mpimg
 from base64 import b64encode
 from fair_research_login.client import NativeClient
 from mdml_client.config import CLIENT_ID
@@ -94,6 +95,26 @@ def _read_image_file(file_name, resize_x=0, resize_y=0, rescale_pixel_intensity=
     img_byte_string = img_b64bytes.decode('utf-8')
     return img_byte_string
 
+def read_any_image(file_name):
+    """
+    Read image from a local file and convert to bytes from sending
+    over the MDML.
+
+
+    Parameters
+    ----------
+    file_name : string
+        file name of the file to be opened
+    
+    Returns
+    -------
+    string
+        String of bytes that can be used as the second argument in 
+        experiment.publish_image()
+    """
+    img = mpimg.imread("/home/jelias/c/Downloads/2019-2/MLLZO_1s1pvp4i12kV_real_scan_2min_00038_00002.tif")
+
+
 def read_image(file_name, resize_x=0, resize_y=0, rescale_pixel_intensity=False):
     """
     Read image from a local file and convert to bytes from sending
@@ -117,7 +138,11 @@ def read_image(file_name, resize_x=0, resize_y=0, rescale_pixel_intensity=False)
         String of bytes that can be used as the second argument in 
         experiment.publish_image()
     """
-    source = cv2.imread(file_name)#, cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH) 
+    try:
+        source = cv2.imread(file_name)#, cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH) 
+    except:
+        source = mpimg.imread(file_name)
+        source = source[...,::-1]
     if resize_x != 0 and resize_y != 0:
         source = cv2.resize(source, (resize_x, resize_y))
     if rescale_pixel_intensity:
@@ -125,12 +150,12 @@ def read_image(file_name, resize_x=0, resize_y=0, rescale_pixel_intensity=False)
         min_val = np.min(source)
         max_val = np.max(source)
         source = (source - min_val) * (255/(max_val - min_val))
-    _, img = cv2.imencode('.jpg', source)
+    _, img = cv2.imencode('.png', source)
     img_b64bytes = b64encode(img)
     img_byte_string = img_b64bytes.decode('utf-8')
     return img_byte_string
 
-def GET_images(image_metadata, experiment_id, host):
+def GET_images(image_metadata, experiment_id, host, verify_cert=True):
     """
     Retrieves images from the MDML using a GET request. Created for
     use in FuncX functions that analyze images.
@@ -144,6 +169,8 @@ def GET_images(image_metadata, experiment_id, host):
         Hostname/IP of the MDML host
     experiment_id : string
         MDML Experiment ID
+    verify_cert : bool
+        Boolean is requests should verify the SSL cert
 
     Returns
     -------
@@ -152,7 +179,7 @@ def GET_images(image_metadata, experiment_id, host):
     """
     imgs = []
     for img in image_metadata:
-        resp = requests.get(f"http://{host}:1880/image?path={img['filepath']}&experiment_id={experiment_id}")
+        resp = requests.get(f"https://{host}:1880/image?path={img['filepath']}&experiment_id={experiment_id}", verify=verify_cert)
         imgs.append(resp.content)
     return imgs
 
@@ -188,7 +215,7 @@ def query_to_pandas(device, query_result, sort=True):
         device_data_pd = device_data_pd.sort_index(axis=1)
     return device_data_pd
 
-def query(query, experiment_id, host):
+def query(query, experiment_id, host, verify_cert=True):
     """
     Query the MDML for an example of the data structure that your query will return. This is aimed at aiding in development of FuncX functions for use with the MDML.
 
@@ -200,6 +227,8 @@ def query(query, experiment_id, host):
         MDML experiment ID for which the data belongs
     host : string
         Host of the MDML instance
+    verify_cert : bool
+        Boolean is requests should verify the SSL cert
     
     Returns
     -------
@@ -207,7 +236,7 @@ def query(query, experiment_id, host):
         Data structure that will be passed to FuncX
     """
     import json
-    resp = requests.get(f"https://{host}:1880/query?query={json.dumps(query)}&experiment_id={experiment_id}")
+    resp = requests.get(f"https://{host}:1880/query?query={json.dumps(query)}&experiment_id={experiment_id}", verify=verify_cert)
     return json.loads(resp.text)
 
 class experiment:
@@ -301,10 +330,12 @@ class experiment:
         """
         Perform a Globus login to acquire auth tokens for FuncX analyses. Must be done before experiment.add_config()
         """
-        scopes = ["https://auth.globus.org/scopes/facd7ccc-c5f4-42aa-916b-a0e270e2c2a9/all"]
+        fx_scope = "https://auth.globus.org/scopes/facd7ccc-c5f4-42aa-916b-a0e270e2c2a9/all"
+        search_scope = "urn:globus:auth:scope:search.api.globus.org:all"
+        scopes = [fx_scope, search_scope, "openid"]
         cli = NativeClient(client_id=CLIENT_ID)
         self.tokens = cli.login(refresh_tokens=True, no_local_server=True, 
-                                no_browser=True, requested_scopes=scopes) 
+                                no_browser=True, requested_scopes=scopes)
 
     def add_config(self, config={}, experiment_run_id="", auto=False):
         """
@@ -444,9 +475,10 @@ class experiment:
         device_id : str
             Unique string identifying the device this data originated from.
             This should correspond with the experiment's configuration
-        data : dict
+        data : dict or list
             Dictionary where keys are the headers for the data device and values are
-            tab delimited strings of data values
+            tab delimited strings of data values. List where each element is a list 
+            of values corresponding to the header name listed for that device.
         timestamp : str
             1 of 3 options: 
                 'none' - influxdb creates timestamp
@@ -459,8 +491,8 @@ class experiment:
         tags : list
             List of variable names (from data) that should be used as tags in InfluxDB (tags should have a finite set of values)
         """
-        if type(data) != dict:
-            print("Error! Data parameter is not a dictionary.")
+        if type(data) != dict and type(data) != list:
+            print("Error! Data parameter is not a dictionary or a list.")
             return
 
         # Creating MQTT topic
@@ -560,7 +592,7 @@ class experiment:
         # Send data via MQTT
         self.client.publish(topic, json.dumps(payload))
 
-    def publish_analysis(self, device_id, function_id, endpoint_id, parameters={}, funcx_callback=None):
+    def publish_analysis(self, device_id, function_id, endpoint_id, parameters={}, funcx_callback=None, trigger=None):
         """
         Publish a message to run an analysis
 
@@ -577,6 +609,9 @@ class experiment:
             Custom parameters to be accessed in the second element of your FuncX data parameter
         funcx_callback : dict
             Dictionary with another set of FuncX params to run with data output from the first FuncX call
+        trigger : list
+            List of device IDs that when data is generated will trigger this 
+            analysis again with the same parameters 
         """
         # Creating MQTT topic
         topic = "MDML/" + self.experiment_id + "/FUNCX/" + device_id
@@ -598,6 +633,11 @@ class experiment:
             # Add to payload
             payload['funcx_callback'] = funcx_callback
 
+        # Add triggers
+        if trigger is not None:
+            assert type(trigger) is list
+            payload['trigger'] = trigger
+
         # Add auth if set
         if self.tokens:
             payload['globus_token'] = self.tokens['funcx_service']['access_token']
@@ -757,7 +797,7 @@ class experiment:
         # Publish it
         self.client.publish(topic, payload)
 
-    def query(self, query):
+    def query(self, query, verify_cert=True):
         """
         Query the MDML for an example of the data structure that your query will return. This is aimed at aiding in development of FuncX functions for use with the MDML.
 
@@ -766,7 +806,8 @@ class experiment:
         ----------
         query : list
             Description of the data to send funcx. See queries format in the documentation on GitHub
-        
+        verify_cert : bool
+            Boolean is requests should verify the SSL cert
 
         Returns
         -------
@@ -774,7 +815,7 @@ class experiment:
             Experiment data from MDML's InfluxDB
         """
         import json
-        resp = requests.get(f"https://{self.host}:1880/query?query={json.dumps(query)}&experiment_id={self.experiment_id}")
+        resp = requests.get(f"https://{self.host}:1880/query?query={json.dumps(query)}&experiment_id={self.experiment_id}", verify=verify_cert)
         return json.loads(resp.text)
 
     def _publish_image_benchmarks(self, device_id, img_byte_string, filename = '', timestamp = 0, size = 0):
@@ -858,15 +899,30 @@ class experiment:
             })
         self.debugger.start()
 
-    def listener(self, func=None):
+    def listener(self, analysis_names=None, callback=None):
         """
         Init a listener to receive any events pertaining to the MDML experiment ID given.
+        
+        
+        Parameters
+        ----------
+        analysis_names : list
+            A list of MDML analyses that you would like to listen to for results.
+        callback : function
+            A callback function to be run on the received messages. 
+            This function must have three parameters. Typically, only the
+            third parameter is used because it contains the topic and message
+            as attributes.
         """
-        if func is None:
+        if callback is None:
             def default_func(client, userdata, message):
                 print(f'TOPIC: {message.topic}\nPAYLOAD: {message.payload.decode("utf-8")}')
-            func = default_func
-        subscribe.callback(func, f"MDML_DEBUG/{self.experiment_id}", hostname=self.host,
+            callback = default_func
+        topics = [f"MDML_DEBUG/{self.experiment_id}"]
+        if analysis_names is not None:
+            for name in analysis_names:
+                topics.append(f"MDML/{self.experiment_id}/RESULTS/{name.upper()}")
+        subscribe.callback(callback, topics, hostname=self.host,
             auth={'username': self.username, 'password': self.password})
 
 
